@@ -42,6 +42,10 @@
 #include "grd-settings.h"
 #include "grd-stream.h"
 
+#ifdef HAVE_NVENC
+#include "grd-rdp-nvenc.h"
+#endif /* HAVE_NVENC */
+
 #define DISCRETE_SCROLL_STEP 10.0
 
 typedef enum _RdpPeerFlag
@@ -124,6 +128,9 @@ struct _GrdSessionRdp
   HANDLE start_event;
   HANDLE stop_event;
 
+  GThread *graphics_thread;
+  GMainContext *graphics_context;
+
   GrdRdpSurface *rdp_surface;
   Pointer *last_pointer;
   GHashTable *pointer_cache;
@@ -143,10 +150,11 @@ struct _GrdSessionRdp
 
 #ifdef HAVE_NVENC
   GrdRdpNvenc *rdp_nvenc;
-#endif /*HAVE_NVENC*/
+#endif /* HAVE_NVENC */
 
   GSource *pending_encode_source;
 
+  GMutex close_session_mutex;
   unsigned int close_session_idle_id;
 
   GrdRdpPipeWireStream *pipewire_stream;
@@ -214,10 +222,10 @@ grd_session_rdp_notify_graphics_pipeline_ready (GrdSessionRdp *session_rdp)
   g_source_set_ready_time (session_rdp->pending_encode_source, 0);
 }
 
-void
-grd_session_rdp_resize_framebuffer (GrdSessionRdp *session_rdp,
-                                    uint32_t       width,
-                                    uint32_t       height)
+static void
+maybe_resize_graphics_output_buffer (GrdSessionRdp *session_rdp,
+                                     uint32_t       width,
+                                     uint32_t       height)
 {
   freerdp_peer *peer = session_rdp->peer;
   rdpSettings *rdp_settings = peer->settings;
@@ -247,6 +255,7 @@ grd_session_rdp_take_buffer (GrdSessionRdp *session_rdp,
   cairo_region_t *region;
 
   g_clear_pointer (&rdp_surface->pending_frame, g_free);
+  maybe_resize_graphics_output_buffer (session_rdp, width, height);
 
   if (is_rdp_peer_flag_set (session_rdp, RDP_PEER_ALL_SURFACES_INVALID))
     {
@@ -564,11 +573,16 @@ grd_session_rdp_hide_pointer (GrdSessionRdp *session_rdp)
 static void
 maybe_queue_close_session_idle (GrdSessionRdp *session_rdp)
 {
+  g_mutex_lock (&session_rdp->close_session_mutex);
   if (session_rdp->close_session_idle_id)
-    return;
+    {
+      g_mutex_unlock (&session_rdp->close_session_mutex);
+      return;
+    }
 
   session_rdp->close_session_idle_id =
     g_idle_add (close_session_idle, session_rdp);
+  g_mutex_unlock (&session_rdp->close_session_mutex);
 
   SetEvent (session_rdp->stop_event);
 }
@@ -788,7 +802,9 @@ rdp_peer_encode_nsc_rect (gpointer data,
 
   g_mutex_lock (thread_pool_context->pending_jobs_mutex);
   --thread_pool_context->pending_job_count;
-  g_cond_signal (thread_pool_context->pending_jobs_cond);
+
+  if (!thread_pool_context->pending_job_count)
+    g_cond_signal (thread_pool_context->pending_jobs_cond);
   g_mutex_unlock (thread_pool_context->pending_jobs_mutex);
 }
 
@@ -962,7 +978,9 @@ rdp_peer_compress_raw_tile (gpointer data,
 
   g_mutex_lock (thread_pool_context->pending_jobs_mutex);
   --thread_pool_context->pending_job_count;
-  g_cond_signal (thread_pool_context->pending_jobs_cond);
+
+  if (!thread_pool_context->pending_job_count)
+    g_cond_signal (thread_pool_context->pending_jobs_cond);
   g_mutex_unlock (thread_pool_context->pending_jobs_mutex);
 }
 
@@ -1603,7 +1621,8 @@ rdp_peer_post_connect (freerdp_peer *peer)
   rdp_settings->PointerCacheSize = MIN (rdp_settings->PointerCacheSize, 100);
 
   session_rdp->rdp_surface = g_malloc0 (sizeof (GrdRdpSurface));
-  session_rdp->rdp_surface->refresh_rate = 30;
+  session_rdp->rdp_surface->refresh_rate = rdp_settings->SupportGraphicsPipeline ? 60
+                                                                                 : 30;
 
   if (rdp_settings->SupportGraphicsPipeline &&
       !rdp_settings->NetworkAutoDetect)
@@ -1625,6 +1644,7 @@ rdp_peer_post_connect (freerdp_peer *peer)
 
       rdp_peer_context->graphics_pipeline =
         grd_rdp_graphics_pipeline_new (session_rdp,
+                                       session_rdp->graphics_context,
                                        rdp_peer_context->vcm,
                                        session_rdp->stop_event,
                                        peer->context,
@@ -1899,6 +1919,27 @@ socket_thread_func (gpointer data)
   return NULL;
 }
 
+static gpointer
+graphics_thread_func (gpointer data)
+{
+  GrdSessionRdp *session_rdp = data;
+
+#ifdef HAVE_NVENC
+  if (session_rdp->rdp_nvenc)
+    grd_rdp_nvenc_push_cuda_context (session_rdp->rdp_nvenc);
+#endif /* HAVE_NVENC */
+
+  while (WaitForSingleObject (session_rdp->stop_event, 0) == WAIT_TIMEOUT)
+    g_main_context_iteration (session_rdp->graphics_context, TRUE);
+
+#ifdef HAVE_NVENC
+  if (session_rdp->rdp_nvenc)
+    grd_rdp_nvenc_pop_cuda_context (session_rdp->rdp_nvenc);
+#endif /* HAVE_NVENC */
+
+  return NULL;
+}
+
 GrdSessionRdp *
 grd_session_rdp_new (GrdRdpServer      *rdp_server,
                      GSocketConnection *connection,
@@ -1938,24 +1979,13 @@ grd_session_rdp_new (GrdRdpServer      *rdp_server,
 #ifdef HAVE_NVENC
   session_rdp->rdp_nvenc = rdp_nvenc;
 #endif /* HAVE_NVENC */
-  session_rdp->start_event = CreateEvent (NULL, TRUE, FALSE, NULL);
-  session_rdp->stop_event = CreateEvent (NULL, TRUE, FALSE, NULL);
 
   session_rdp->socket_thread = g_thread_new ("RDP socket thread",
                                              socket_thread_func,
                                              session_rdp);
-  if (!session_rdp->socket_thread)
-    {
-      g_warning ("Failed to create socket thread");
-
-      g_clear_pointer (&session_rdp->stop_event, CloseHandle);
-      g_clear_pointer (&session_rdp->start_event, CloseHandle);
-      g_object_unref (connection);
-      g_free (password);
-      g_free (username);
-
-      return NULL;
-    }
+  session_rdp->graphics_thread = g_thread_new ("RDP graphics thread",
+                                               graphics_thread_func,
+                                               session_rdp);
 
   init_rdp_session (session_rdp, username, password);
 
@@ -2006,6 +2036,15 @@ grd_session_rdp_stop (GrdSession *session)
         rdp_peer_context->network_autodetection);
     }
 
+  if (session_rdp->graphics_thread)
+    {
+      g_assert (session_rdp->graphics_context);
+      g_assert (WaitForSingleObject (session_rdp->stop_event, 0) != WAIT_TIMEOUT);
+
+      g_main_context_wakeup (session_rdp->graphics_context);
+      g_clear_pointer (&session_rdp->graphics_thread, g_thread_join);
+    }
+
   g_clear_object (&session_rdp->pipewire_stream);
 
   g_clear_object (&rdp_peer_context->clipboard_rdp);
@@ -2042,9 +2081,6 @@ grd_session_rdp_stop (GrdSession *session)
                                clear_pointer_bitmap,
                                NULL);
 
-  g_clear_pointer (&session_rdp->stop_event, CloseHandle);
-  g_clear_pointer (&session_rdp->start_event, CloseHandle);
-
   g_clear_handle_id (&session_rdp->close_session_idle_id, g_source_remove);
 }
 
@@ -2086,6 +2122,7 @@ grd_session_rdp_stream_ready (GrdSession *session,
                               GrdStream  *stream)
 {
   GrdSessionRdp *session_rdp = GRD_SESSION_RDP (session);
+  GMainContext *graphics_context = session_rdp->graphics_context;
   uint32_t pipewire_node_id;
   uint16_t refresh_rate;
   g_autoptr (GError) error = NULL;
@@ -2093,6 +2130,7 @@ grd_session_rdp_stream_ready (GrdSession *session,
   pipewire_node_id = grd_stream_get_pipewire_node_id (stream);
   refresh_rate = session_rdp->rdp_surface->refresh_rate;
   session_rdp->pipewire_stream = grd_rdp_pipewire_stream_new (session_rdp,
+                                                              graphics_context,
                                                               pipewire_node_id,
                                                               refresh_rate,
                                                               &error);
@@ -2138,10 +2176,16 @@ grd_session_rdp_dispose (GObject *object)
       g_clear_pointer (&session_rdp->pending_encode_source, g_source_unref);
     }
 
+  g_assert (!session_rdp->graphics_thread);
+  g_clear_pointer (&session_rdp->graphics_context, g_main_context_unref);
+
   g_clear_pointer (&session_rdp->pressed_unicode_keys, g_hash_table_unref);
   g_clear_pointer (&session_rdp->pressed_keys, g_hash_table_unref);
   g_clear_pointer (&session_rdp->pointer_cache, g_hash_table_unref);
 
+  g_clear_pointer (&session_rdp->stop_event, CloseHandle);
+  g_clear_pointer (&session_rdp->start_event, CloseHandle);
+
   G_OBJECT_CLASS (grd_session_rdp_parent_class)->dispose (object);
 }
 
@@ -2201,6 +2245,9 @@ static GSourceFuncs pending_encode_source_funcs =
 static void
 grd_session_rdp_init (GrdSessionRdp *session_rdp)
 {
+  session_rdp->start_event = CreateEvent (NULL, TRUE, FALSE, NULL);
+  session_rdp->stop_event = CreateEvent (NULL, TRUE, FALSE, NULL);
+
   session_rdp->pointer_cache = g_hash_table_new (NULL, are_pointer_bitmaps_equal);
   session_rdp->pressed_keys = g_hash_table_new (NULL, NULL);
   session_rdp->pressed_unicode_keys = g_hash_table_new (NULL, NULL);
@@ -2208,15 +2255,19 @@ grd_session_rdp_init (GrdSessionRdp *session_rdp)
   g_cond_init (&session_rdp->pending_jobs_cond);
   g_mutex_init (&session_rdp->pending_jobs_mutex);
   g_mutex_init (&session_rdp->rdp_flags_mutex);
+  g_mutex_init (&session_rdp->close_session_mutex);
 
   session_rdp->rdp_event_queue = grd_rdp_event_queue_new (session_rdp);
 
+  session_rdp->graphics_context = g_main_context_new ();
+
   session_rdp->pending_encode_source = g_source_new (&pending_encode_source_funcs,
                                                      sizeof (GSource));
   g_source_set_callback (session_rdp->pending_encode_source,
                          encode_pending_frames, session_rdp, NULL);
   g_source_set_ready_time (session_rdp->pending_encode_source, -1);
-  g_source_attach (session_rdp->pending_encode_source, NULL);
+  g_source_attach (session_rdp->pending_encode_source,
+                   session_rdp->graphics_context);
 }
 
 static void
